@@ -20,22 +20,75 @@ module Mixlib
   class ShellOut
     module Unix
 
+      # "1.8.7" as a frozen string. We use this with a hack that disables GC to
+      # avoid segfaults on Ruby 1.8.7, so we need to allocate the fewest
+      # objects we possibly can.
+      ONE_DOT_EIGHT_DOT_SEVEN = "1.8.7".freeze
+
       # Option validation that is unix specific
       def validate_options(opts)
         # No options to validate, raise exceptions here if needed
       end
 
+      # Whether we're simulating a login shell
+      def using_login?
+        return login && user
+      end
+
+      # Helper method for sgids
+      def all_seconderies
+        ret = []
+        Etc.endgrent
+        while ( g = Etc.getgrent ) do
+          ret << g
+        end
+        Etc.endgrent
+        return ret
+      end
+
+      # The secondary groups that the subprocess will switch to.
+      # Currently valid only if login is used, and is set
+      # to the user's secondary groups
+      def sgids
+        return nil unless using_login?
+        user_name = Etc.getpwuid(uid).name
+        all_seconderies.select{|g| g.mem.include?(user_name)}.map{|g|g.gid}
+      end
+
+      # The environment variables that are deduced from simulating logon
+      # Only valid if login is used
+      def logon_environment
+        return {} unless using_login?
+        entry = Etc.getpwuid(uid)
+        # According to `man su`, the set fields are:
+        #  $HOME, $SHELL, $USER, $LOGNAME, $PATH, and $IFS
+        # Values are copied from "shadow" package in Ubuntu 14.10
+        {'HOME'=>entry.dir, 'SHELL'=>entry.shell, 'USER'=>entry.name, 'LOGNAME'=>entry.name, 'PATH'=>'/sbin:/bin:/usr/sbin:/usr/bin', 'IFS'=>"\t\n"}
+      end
+
+      # Merges the two environments for the process
+      def process_environment
+        logon_environment.merge(self.environment)
+      end
+
       # Run the command, writing the command's standard out and standard error
       # to +stdout+ and +stderr+, and saving its exit status object to +status+
       # === Returns
       # returns   +self+; +stdout+, +stderr+, +status+, and +exitstatus+ will be
-      # populated with results of the command
+      # populated with results of the command.
       # === Raises
       # * Errno::EACCES  when you are not privileged to execute the command
       # * Errno::ENOENT  when the command is not available on the system (or not
       #   in the current $PATH)
       # * Chef::Exceptions::CommandTimeout  when the command does not complete
-      #   within +timeout+ seconds (default: 600s)
+      #   within +timeout+ seconds (default: 600s). When this happens, ShellOut
+      #   will send a TERM and then KILL to the entire process group to ensure
+      #   that any grandchild processes are terminated. If the invocation of
+      #   the child process spawned multiple child processes (which commonly
+      #   happens if the command is passed as a single string to be interpreted
+      #   by bin/sh, and bin/sh is not bash), the exit status object may not
+      #   contain the correct exit code of the process (of course there is no
+      #   exit code if the command is killed by SIGKILL, also).
       def run_command
         @child_pid = fork_subprocess
         @reaped = false
@@ -43,15 +96,16 @@ module Mixlib
         configure_parent_process_file_descriptors
 
         # Ruby 1.8.7 and 1.8.6 from mid 2009 try to allocate objects during GC
-        # when calling IO.select and IO#read. Some OS Vendors are not interested
-        # in updating their ruby packages (Apple, *cough*) and we *have to*
-        # make it work. So I give you this epic hack:
-        GC.disable
+        # when calling IO.select and IO#read. Disabling GC works around the
+        # segfault, but obviously it's a bad workaround. We no longer support
+        # 1.8.6 so we only need this hack for 1.8.7.
+        GC.disable if RUBY_VERSION == ONE_DOT_EIGHT_DOT_SEVEN
 
         # CHEF-3390: Marshall.load on Ruby < 1.8.7p369 also has a GC bug related
         # to Marshall.load, so try disabling GC first.
         propagate_pre_exec_failure
 
+        @status = nil
         @result = nil
         @execution_time = 0
 
@@ -96,8 +150,8 @@ module Mixlib
 
       def set_user
         if user
-          Process.euid = uid
           Process.uid = uid
+          Process.euid = uid
         end
       end
 
@@ -108,8 +162,15 @@ module Mixlib
         end
       end
 
+      def set_secondarygroups
+        if sgids
+          Process.groups = sgids
+        end
+      end
+
       def set_environment
-        environment.each do |env_var,value|
+        # user-set variables should override the login ones
+        process_environment.each do |env_var,value|
           ENV[env_var] = value
         end
       end
@@ -122,6 +183,12 @@ module Mixlib
         Dir.chdir(cwd) if cwd
       end
 
+      # Since we call setsid the child_pgid will be the child_pid, set to negative here
+      # so it can be directly used in arguments to kill, wait, etc.
+      def child_pgid
+        -@child_pid
+      end
+
       def initialize_ipc
         @stdin_pipe, @stdout_pipe, @stderr_pipe, @process_status_pipe = IO.pipe, IO.pipe, IO.pipe, IO.pipe
         @process_status_pipe.last.fcntl(Fcntl::F_SETFD, Fcntl::FD_CLOEXEC)
@@ -176,31 +243,6 @@ module Mixlib
         STDIN.sync = true if input
       end
 
-      # When a new process is started with chef, it shares the file
-      # descriptors of the parent. We clean the file descriptors
-      # coming from the parent to prevent unintended locking if parent
-      # is killed.
-      # NOTE: After some discussions we've decided to iterate on file
-      # descriptors upto 256. We believe this  is a reasonable upper
-      # limit in a chef environment. If we have issues in the future this
-      # number could be made to be configurable or updated based on
-      # the ulimit based on platform.
-      def clean_parent_file_descriptors
-        # Don't clean $stdin, $stdout, $stderr, process_status_pipe.
-        3.upto(256) do |n|
-          # We are checking the fd for error pipe before attempting to
-          # create a file because error pipe will auto close when we
-          # try to create a file since it's set to CLOEXEC.
-          if n != @process_status_pipe.last.to_i
-            begin
-              fd = File.for_fd(n)
-              fd.close if fd
-            rescue
-            end
-          end
-        end
-      end
-
       def configure_parent_process_file_descriptors
         # Close the sides of the pipes we don't care about
         stdin_pipe.first.close
@@ -219,7 +261,7 @@ module Mixlib
       # Some patch levels of ruby in wide use (in particular the ruby 1.8.6 on OSX)
       # segfault when you IO.select a pipe that's reached eof. Weak sauce.
       def open_pipes
-        @open_pipes ||= [child_stdout, child_stderr]
+        @open_pipes ||= [child_stdout, child_stderr, child_process_status]
       end
 
       # Keep this unbuffered for now
@@ -231,11 +273,10 @@ module Mixlib
 
       def attempt_buffer_read
         ready = IO.select(open_pipes, nil, nil, READ_WAIT_TIME)
-        if ready && ready.first.include?(child_stdout)
-          read_stdout_to_buffer
-        end
-        if ready && ready.first.include?(child_stderr)
-          read_stderr_to_buffer
+        if ready
+          read_stdout_to_buffer if ready.first.include?(child_stdout)
+          read_stderr_to_buffer if ready.first.include?(child_stderr)
+          read_process_status_to_buffer if ready.first.include?(child_process_status)
         end
         ready
       end
@@ -243,7 +284,7 @@ module Mixlib
       def read_stdout_to_buffer
         while chunk = child_stdout.read_nonblock(READ_SIZE)
           @stdout << chunk
-          @live_stream << chunk if @live_stream
+          @live_stdout << chunk if @live_stdout
         end
       rescue Errno::EAGAIN
       rescue EOFError
@@ -253,20 +294,40 @@ module Mixlib
       def read_stderr_to_buffer
         while chunk = child_stderr.read_nonblock(READ_SIZE)
           @stderr << chunk
+          @live_stderr << chunk if @live_stderr
         end
       rescue Errno::EAGAIN
       rescue EOFError
         open_pipes.delete(child_stderr)
       end
 
+      def read_process_status_to_buffer
+        while chunk = child_process_status.read_nonblock(READ_SIZE)
+          @process_status << chunk
+        end
+      rescue Errno::EAGAIN
+      rescue EOFError
+        open_pipes.delete(child_process_status)
+      end
+
       def fork_subprocess
         initialize_ipc
 
         fork do
+          # Child processes may themselves fork off children. A common case
+          # is when the command is given as a single string (instead of
+          # command name plus Array of arguments) and /bin/sh does not
+          # support the "ONESHOT" optimization (where sh -c does exec without
+          # forking). To support cleaning up all the children, we need to
+          # ensure they're in a unique process group.
+          #
+          # We use setsid here to abandon our controlling tty and get a new session
+          # and process group that are set to the pid of the child process.
+          Process.setsid
+
           configure_subprocess_file_descriptors
 
-          clean_parent_file_descriptors
-
+          set_secondarygroups
           set_group
           set_user
           set_environment
@@ -274,7 +335,7 @@ module Mixlib
           set_cwd
 
           begin
-            command.kind_of?(Array) ? exec(*command) : exec(command)
+            command.kind_of?(Array) ? exec(*command, :close_others=>true) : exec(command, :close_others=>true)
 
             raise 'forty-two' # Should never get here
           rescue Exception => e
@@ -291,25 +352,26 @@ module Mixlib
       # assume everything went well.
       def propagate_pre_exec_failure
         begin
-          e = Marshal.load child_process_status
+          attempt_buffer_read until child_process_status.eof?
+          e = Marshal.load(@process_status)
           raise(Exception === e ? e : "unknown failure: #{e.inspect}")
-        rescue EOFError # If we get an EOF error, then the exec was successful
+        rescue ArgumentError # If we get an ArgumentError error, then the exec was successful
           true
         ensure
           child_process_status.close
+          open_pipes.delete(child_process_status)
         end
       end
 
       def reap_errant_child
         return if attempt_reap
-        @terminate_reason = "Command execeded allowed execution time, killed by TERM signal."
-        logger.error("Command execeded allowed execution time, sending TERM") if logger
-        Process.kill(:TERM, @child_pid)
+        @terminate_reason = "Command exceeded allowed execution time, process terminated"
+        logger.error("Command exceeded allowed execution time, sending TERM") if logger
+        Process.kill(:TERM, child_pgid)
         sleep 3
-        return if attempt_reap
-        @terminate_reason = "Command execeded allowed execution time, did not respond to TERM. Killed by KILL signal."
-        logger.error("Command did not exit from TERM, sending KILL") if logger
-        Process.kill(:KILL, @child_pid)
+        attempt_reap
+        logger.error("Command exceeded allowed execution time, sending KILL") if logger
+        Process.kill(:KILL, child_pgid)
         reap
 
         # Should not hit this but it's possible if something is calling waitall
@@ -323,12 +385,22 @@ module Mixlib
         @child_pid && !@reaped
       end
 
+      # Unconditionally reap the child process. This is used in scenarios where
+      # we can be confident the child will exit quickly, and has not spawned
+      # and grandchild processes.
       def reap
         results = Process.waitpid2(@child_pid)
         @reaped = true
         @status = results.last
+      rescue Errno::ECHILD
+        # When cleaning up timed-out processes, we might send SIGKILL to the
+        # whole process group after we've cleaned up the direct child. In that
+        # case the grandchildren will have been adopted by init so we can't
+        # reap them even if we wanted to (we don't).
+        nil
       end
 
+      # Try to reap the child process but don't block if it isn't dead yet.
       def attempt_reap
         if results = Process.waitpid2(@child_pid, Process::WNOHANG)
           @reaped = true
